"""Sacramento Municipal Utility District (SMUD)."""

#
# SMUD is a community owned, local monopoly power company in Sacramento, California.
#
# https://smud.org
# https://en.wikipedia.org/wiki/Sacramento_Municipal_Utility_District
# https://www.reddit.com/r/homeassistant/comments/10wnoox/smud_energy_smart_meter_integration/
#
# The web UI for power usage is at:
# https://myaccount.smud.org/manage/opowerresidential/energyusage
#
# OKTA is an enterprise identity/SSO provider
# https://www.okta.com/
#
# To filter FireFox network inspector when logging in:
# `-regexp:woff|beacon|appdynamics|fonts|css|hotjar|usabilla|chat|ico|analytics|promo|nextprofilequestion|trackEvent
# -mime-type:image`
#
# Test with:
# `python src/demo.py --utility smud --username mysmudloginemail@example.com --password "mypassword" -v`

from html.parser import HTMLParser
import logging
from typing import Optional
from urllib.parse import parse_qs

from aiohttp import ClientResponse, ClientSession
from aiohttp.client_exceptions import ClientResponseError
from yarl import URL

import opower

from ..const import USER_AGENT
from .base import UtilityBase

_LOGGER = logging.getLogger(__file__)


class SMUDLoginParser(HTMLParser):
    """HTML parser to extract login verification token from SMUD Login page."""

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self.verification_token: Optional[str] = None
        self.ocis_req_sp: Optional[str] = None
        self.relay_state: Optional[str] = None

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Try to extract the verification token from the login input."""
        if tag == "input" and ("name", "__RequestVerificationToken") in attrs:
            _, token = next(filter(lambda attr: attr[0] == "value", attrs))
            if token is None:
                return
            _LOGGER.debug(
                "SMUD self verify token: %s...%s (%d characters)",
                token[0:5],
                token[-5:],
                len(token),
            )
            self.verification_token = token
        """Try to extract the OCIS_REQ_SP input value from the identity.oraclecloud.com sso HTML"""
        if tag == "input" and ("name", "OCIS_REQ_SP") in attrs:
            _, token = next(filter(lambda attr: attr[0] == "value", attrs))
            if token is None:
                return
            _LOGGER.debug(
                "OCIS_REQ_SP self verify token: %s...%s (%d characters)",
                token[0:5],
                token[-5:],
                len(token),
            )
            self.ocis_req_sp = token
        """Try to extract the RelayState input value from the smud.okta.com HTML"""
        if tag == "input" and ("name", "RelayState") in attrs:
            _, token = next(filter(lambda attr: attr[0] == "value", attrs))
            if token is None:
                return
            _LOGGER.debug(
                "RelayState value: %s...%s (%d characters)",
                token[0:5],
                token[-5:],
                len(token),
            )
            self.relay_state = token


class SMUDOktaResponseSamlResponseValueParser(HTMLParser):
    """HTML parser to extract SAMLResponse token from OKTA response for Opower SSO."""

    # <input name="SAMLResponse" type="hidden" value="..."/>
    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Try to extract the SAMLResponse value."""
        if tag == "input":
            for name, value in attrs:
                if name == "name" and value == "SAMLResponse":
                    self.saml_response = attrs[2][1]


class SMUD(UtilityBase):
    """Sacramento Municipal Utility District (SMUD)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Sacramento Municipal Utility District (SMUD)"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "smud"

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/Los_Angeles"

    @staticmethod
    async def async_login(
        session: ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],  # Not used by SMUD.
    ) -> None:
        """Login to the utility website and authorize opower."""
        # If we already have a cookie, return early if it is valid.
        if (
            len(session.cookie_jar.filter_cookies(URL("https://smud.opower.com/ei")))
            > 0
        ):
            try:
                async with session.get(
                    "https://smud.opower.com/ei/edge/apis/multi-account-v1/cws/smud/customers",
                    headers={"User-Agent": USER_AGENT},
                    raise_for_status=True,
                ):
                    return
            except ClientResponseError:
                _LOGGER.debug("Failed to login to SMUD with existing cookies")
                session.cookie_jar.clear()
                pass

        smud_login_page_url = "https://myaccount.smud.org/"

        _LOGGER.debug("Fetching SMUD login page: %s", smud_login_page_url)

        myaccount_response = await session.get(
            smud_login_page_url,
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        )

        await SMUD.log_response(myaccount_response, session)

        # Parse the verification token which will be used during login.
        # NB: Although the response cookies contain a `__RequestVerificationToken`, it does not match
        # the one contained in the HTML and results in a 500 server error from SMUD during login.
        login_parser = SMUDLoginParser()
        login_parser.feed(await myaccount_response.text())
        requestVerificationToken = login_parser.verification_token

        _LOGGER.debug("Performing SMUD login to: %s", smud_login_page_url)

        # Do the SMUD login which will set cookies in the session.
        login_response = await session.post(
            smud_login_page_url,
            data={
                "Lang": "en",
                "UserID": username,
                "Password": password,
                "__RequestVerificationToken": requestVerificationToken,
            },
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        )

        await SMUD.log_response(login_response, session)

        login_response_body = await login_response.text()
        if "could not be authenticated" in login_response_body:
            raise opower.InvalidAuth

        smud_energyusage_page_url = (
            "https://myaccount.smud.org/manage/opowerresidential/energyusage"
        )

        _LOGGER.debug("Opening SMUD energy usage page: %s", smud_energyusage_page_url)

        # Visit the Energy Usage page
        energyusage_response = await session.get(
            smud_energyusage_page_url,
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        )

        await SMUD.log_response(energyusage_response, session)

        okta_login_2_url = SMUD.get_okta_url_from_response_redirect(
            energyusage_response
        )

        _LOGGER.debug("Fetching second OKTA login page: %s", okta_login_2_url)

        smud_okta_response = await session.get(
            okta_login_2_url,
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        )

        await SMUD.log_response(smud_okta_response, session)

        parser = SMUDOktaResponseSamlResponseValueParser()
        parser.feed(await smud_okta_response.text())
        saml_response = parser.saml_response
        assert saml_response

        _LOGGER.debug(
            "Parsed SAMLResponse: %s...%s (%d characters)",
            saml_response[0:5],
            saml_response[-5:],
            len(saml_response),
        )

        # This step is done in the web browser but doesn't seem to matter here.
        smud_ssotransition_url = "https://myaccount.smud.org/signin/ssotransition"

        _LOGGER.debug("Fetching SMUD ssotransition page: %s", smud_ssotransition_url)
        smud_ssotransition_response = await session.get(
            smud_ssotransition_url,
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        )

        await SMUD.log_response(smud_ssotransition_response, session)

        # This is the action of the #appForm form in the smud_okta_response HTML.
        opower_sso_url = "https://idcs-8d184356671642c58ea38b42e6420ed2.identity.oraclecloud.com/fed/v1/sp/sso"

        _LOGGER.debug("POSTing opower sso page with SAMLResponse: %s", opower_sso_url)

        opower_sso_response = await session.post(
            opower_sso_url,
            data={
                "SAMLResponse": saml_response,
                "RelayState": "https://smud.opower.com/ei/app/myEnergyUse",
            },
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
            allow_redirects=True,
        )
        await SMUD.log_response(opower_sso_response, session)

        login_parser.feed(await opower_sso_response.text())
        ocis_req_sp = login_parser.ocis_req_sp

        identity_oraclecloud_login_url = "https://idcs-8d184356671642c58ea38b42e6420ed2.identity.oraclecloud.com/sso/v1/user/login"

        _LOGGER.debug(
            "POSTing opower sso login page with OCIS_REQ_SP: %s",
            identity_oraclecloud_login_url,
        )

        identity_oraclecloud_login_response = await session.post(
            identity_oraclecloud_login_url,
            data={
                "OCIS_REQ_SP": ocis_req_sp,
            },
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
            allow_redirects=True,
            max_redirects=10,
        )

        await SMUD.log_response(identity_oraclecloud_login_response, session)

        okta_saml_request_url = identity_oraclecloud_login_response.real_url
        _LOGGER.debug(
            "RESPONSE from opower sso login page with OCIS_REQ_SP, RelayState: %s",
            okta_saml_request_url,
        )

        return

    @classmethod
    def get_okta_url_from_response_redirect(
        cls, energyusage_response: ClientResponse
    ) -> str:
        """Get the OKTA URL to open next from the last redirect of the previous response."""
        # https://smud.okta.com/login/sessionCookieRedirect
        #   ?token=20111...6QJMn
        #   &redirectUrl=https://smud.okta.com/app/sacramentomunicipalutilitydistrict_opower_1/exk2i...cF0x7/sso/saml
        #       ?RelayState=https://smud.opower.com/ei/app/myEnergyUse
        energyUsageResponseRedirectedFinalUrl = energyusage_response.history[-1].url

        query_parts = parse_qs(energyUsageResponseRedirectedFinalUrl.query_string)

        return query_parts["redirectUrl"][0]

    # Store cookies so we can log what is new after each request.
    cookies: dict[str, list[str]] = {}

    @staticmethod
    async def log_response(response: ClientResponse, session: ClientSession) -> None:
        """Log any redirects and new cookies. Log full HTML when -vv is set."""
        host = response.host  # Is this the request URL or the final redirected url?

        redirects = [r.url for r in response.history]
        if len(redirects) > 1:
            _LOGGER.debug("Performed %d redirects", len(redirects) - 1)
            for redirect in redirects[1:]:
                _LOGGER.debug("-> %s", redirect.__str__())

        if len(session.cookie_jar.filter_cookies(response.url)) > 0:
            response_cookie_names = list(
                session.cookie_jar.filter_cookies(response.url).keys()
            )
            last_cookie_names = SMUD.cookies.get(host, [])
            response_new_cookie_names = set(response_cookie_names) - set(
                last_cookie_names
            )

            if len(response_new_cookie_names) > 0:
                _LOGGER.debug(
                    "Set new cookies: `%s`", "`, `".join(response_new_cookie_names)
                )

                SMUD.cookies[host] = last_cookie_names + response_cookie_names

        response_html = await response.text()
        _LOGGER.log(logging.DEBUG - 1, "Response %s:", response.url)
        _LOGGER.log(logging.DEBUG - 1, response_html)



"""Exceptions."""

from typing import Optional


class CannotConnect(Exception):
    """Error to indicate we cannot connect."""


class InvalidAuth(Exception):
    """Error to indicate there is invalid auth."""


class ApiException(Exception):
    """Raised during problems talking to the API."""

    def __init__(
        self,
        message: str,
        url: str,
        status: Optional[int] = None,
        response_text: Optional[str] = None,
    ) -> None:
        """Initialize the exception."""
        super().__init__(message)
        self.url = url
        self.status = status
        self.response_text = response_text

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        parts = [super().__str__()]
        parts.append(f"URL: {self.url}")
        if self.status is not None:
            parts.append(f"Status: {self.status}")
        if self.response_text is not None:
            parts.append(f"Response: {self.response_text}")
        return "\n".join(parts)



"""AEP Ohio."""

from .aepbase import AEPBase
from .base import UtilityBase


class AEPOHio(AEPBase, UtilityBase):
    """AEP Ohio."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "AEP Ohio"

    @staticmethod
    def hostname() -> str:
        """Return the hostname for login."""
        return "aepohio.com"



"""Southwestern Electric Power Company (SWEPCO)."""

from .aepbase import AEPBase
from .base import UtilityBase


class SWEPCO(AEPBase, UtilityBase):
    """Southwestern Electric Power Company (SWEPCO)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Southwestern Electric Power Company (SWEPCO)"

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/Chicago"

    @staticmethod
    def hostname() -> str:
        """Return the hostname for login."""
        return "swepco.com"



"""City of Austin Utilities."""

from typing import Optional
from urllib.parse import parse_qs, urlparse

import aiohttp
from yarl import URL

from ..const import USER_AGENT
from ..exceptions import InvalidAuth
from .base import UtilityBase
from .helpers import get_form_action_url_and_hidden_inputs


class COAUtilities(UtilityBase):
    """City of Austin Utilities."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "City of Austin Utilities"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "coa"

    @staticmethod
    def timezone() -> str:
        """Return the timezone.

        Should match the siteTimeZoneId of the API responses.
        """
        return "America/Chicago"

    @staticmethod
    def is_dss() -> bool:
        """Check if Utility using DSS version of the portal."""
        return True

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> Optional[str]:
        """Login to the utility website."""
        # Get cookies
        await session.get(
            "https://coautilities.com/wps/wcm/connect/occ/coa/home",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        )

        # Auth using username and password on coautilities
        url = (
            "https://coautilities.com/pkmslogin.form?/isam/sps/OPowerIDP_DSS/saml20/logininitial?"
            "RequestBinding=HTTPPost&"
            "NameIdFormat=email&"
            "PartnerId=opower-coa-dss-webUser&"
            "Target=https://dss-coa.opower.com"
        )

        async with session.post(
            url,
            headers={"User-Agent": USER_AGENT},
            data={
                "username": username,
                "password": password,
                "login-form-type": "pwd",
            },
            raise_for_status=True,
        ) as response:
            await response.text()
            if "PD-S-SESSION-ID-PCOAUT" not in session.cookie_jar.filter_cookies(
                URL("https://coautilities.com")
            ):
                raise InvalidAuth("Username/Password are invalid")

        # Getting SAML Request from opower
        url = (
            "https://sso.opower.com/sp/startSSO.ping?"
            "PartnerIdpId=https://coautilities.com/isam/sps/OPowerIDP_DSS/saml20&"
            "TargetResource=https%3A%2F%2Fdss-coa.opower.com%2Fwebcenter%2Fedge%2Fapis%2Fidentity-management-v1%2Fcws"
            "%2Fv1%2Fauth%2Fcoa%2Fsaml%2Flogin%2Fcallback%3FsuccessUrl%3Dhttps%253A%252F%252Fdss-coa.opower.com%252Fdss"
            "%252Flogin-success%253Ftoken%253D%2525s%2526nextPathname%253DL2Rzcy8%253D%26failureUrl%3Dhttps%253A%252F"
            "%252Fdss-coa.opower.com%252Fdss%252Flogin-error%253Freason%253D%2525s"
        )

        async with session.post(url, raise_for_status=True) as response:
            html = await response.text()
            action_url, hidden_inputs = get_form_action_url_and_hidden_inputs(html)
            assert set(hidden_inputs.keys()) == {"RelayState", "SAMLRequest"}

        # Getting SAML Response from coautilities
        headers = {
            "Referer": "https://sso.opower.com/",
            "User-Agent": USER_AGENT,
        }
        async with session.post(
            action_url,
            headers=headers,
            data=hidden_inputs,
            raise_for_status=True,
        ) as response:
            html = await response.text()
            action_url, hidden_inputs = get_form_action_url_and_hidden_inputs(html)
            assert set(hidden_inputs.keys()) == {"RelayState", "SAMLResponse"}

        # Getting Open Token from opower
        async with session.post(
            action_url,
            headers={"User-Agent": USER_AGENT},
            data=hidden_inputs,
            raise_for_status=True,
        ) as response:
            html = await response.text()
            action_url, hidden_inputs = get_form_action_url_and_hidden_inputs(html)
            assert set(hidden_inputs.keys()) == {"opentoken"}

        session.cookie_jar.update_cookies({"dssPortalCW": "1"})

        # Getting success token
        async with session.post(
            action_url,
            headers={"User-Agent": USER_AGENT},
            data=hidden_inputs,
            allow_redirects=False,
            raise_for_status=True,
        ) as response:
            await response.text()
            parsed_url = urlparse(response.headers["Location"])
            parsed_query = parse_qs(parsed_url.query)
            assert "token" in parsed_query
            token = parsed_query["token"][0]

        # Finally exchange this token to Auth token
        async with session.post(
            "https://dss-coa.opower.com"
            "/webcenter/edge/apis/identity-management-v1/cws/v1/auth/coa/saml/ott/confirm",
            headers={"User-Agent": USER_AGENT},
            data={"token": token},
            raise_for_status=True,
        ) as response:
            content = await response.json()
            return str(content["sessionToken"])



"""Helper functions."""

import base64
import re

import aiohttp
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa

from ..const import USER_AGENT


def get_form_action_url_and_hidden_inputs(html: str) -> tuple[str, dict[str, str]]:
    """Return the URL and hidden inputs from the single form in a page."""
    match = re.search(r'action="([^"]*)"', html)
    if not match:
        return "", {}
    action_url = match.group(1)
    inputs = {}
    for match in re.finditer(
        r'input\s*type="hidden"\s*name="([^"]*)"\s*value="([^"]*)"', html
    ):
        inputs[match.group(1)] = match.group(2)
    return action_url, inputs


async def async_auth_saml(session: aiohttp.ClientSession, url: str) -> None:
    """Authenticate with Opower using SAML."""
    # Fetch the URL on the utility website to get RelayState and SAMLResponse.
    async with session.get(
        url,
        headers={"User-Agent": USER_AGENT},
        raise_for_status=True,
    ) as resp:
        result = await resp.text()
    action_url, hidden_inputs = get_form_action_url_and_hidden_inputs(result)
    assert action_url.endswith(".opower.com/sp/ACS.saml2")
    assert set(hidden_inputs.keys()) == {"RelayState", "SAMLResponse"}

    # Pass them to opower.com/sp/ACS.saml2 to get opentoken.
    async with session.post(
        action_url,
        data=hidden_inputs,
        headers={"User-Agent": USER_AGENT},
        raise_for_status=True,
    ) as resp:
        result = await resp.text()
    action_url, hidden_inputs = get_form_action_url_and_hidden_inputs(result)
    if action_url == "":
        return
    assert set(hidden_inputs.keys()) == {"opentoken"}

    # Pass it back to the utility website.
    async with session.post(
        action_url,
        data=hidden_inputs,
        headers={"User-Agent": USER_AGENT},
        raise_for_status=True,
    ) as resp:
        pass


def js_encrypt(pub_key: str, text: str) -> str:
    """JSEncrypt-like encryption function using cryptography."""
    # Load the public key
    rsakey = serialization.load_pem_public_key(pub_key.encode())
    # Check if the key is RSA before encrypting
    if isinstance(rsakey, rsa.RSAPublicKey):
        # Encrypt the text using RSA and PKCS1v15 padding
        cipher_text = rsakey.encrypt(text.encode(), padding.PKCS1v15())

        # Encode the encrypted text in base64
        cipher_text_base64 = base64.b64encode(cipher_text)

        return cipher_text_base64.decode()
    else:
        raise ConnectionError("Could not find public key to and encrypt password.")



"""Potomac Electric Power Company (Pepco)."""

from .base import UtilityBase
from .exelon import Exelon


class Pepco(Exelon, UtilityBase):
    """Potomac Electric Power Company (Pepco)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Potomac Electric Power Company (Pepco)"

    @staticmethod
    def primary_subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "pep"

    @staticmethod
    def secondary_subdomain() -> str:
        """Return the opower.com secondary subdomain for this utility."""
        return "pepd"

    @staticmethod
    def login_domain() -> str:
        """Return the domain that hosts the login page."""
        return "secure.pepco.com"



"""Base class that each utility needs to extend."""

from typing import Any, Optional

import aiohttp


class UtilityBase:
    """Base class that each utility needs to extend."""

    subclasses: list[type["UtilityBase"]] = []

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Keep track of all subclass implementations."""
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        raise NotImplementedError

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        raise NotImplementedError

    @staticmethod
    def timezone() -> str:
        """Return the timezone.

        Should match the siteTimeZoneId of the API responses.
        """
        raise NotImplementedError

    @staticmethod
    def accepts_mfa() -> bool:
        """Check if Utility implementations supports MFA."""
        return False

    @staticmethod
    def is_dss() -> bool:
        """Check if Utility using DSS version of the portal."""
        return False

    @classmethod
    def utilitycode(cls) -> str:
        """Return the utilitycode identifier for the utility."""
        return cls.subdomain()

    @staticmethod
    def supports_realtime_usage() -> bool:
        """Check if Utility supports realtime usage reads."""
        return False

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> Optional[str]:
        """Login to the utility website.

        Return the Opower access token or None if this function authorizes with Opower in other ways.

        :raises InvalidAuth: if login information is incorrect
        """
        raise NotImplementedError



"""Pacific Gas & Electric (PG&E)."""

from typing import Optional
import urllib.parse

import aiohttp

from ..const import USER_AGENT
from ..exceptions import InvalidAuth
from .base import UtilityBase
from .helpers import async_auth_saml


class PGE(UtilityBase):
    """Pacific Gas & Electric (PG&E)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Pacific Gas and Electric Company (PG&E)"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "pge"

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/Los_Angeles"

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> None:
        """Login to the utility website."""
        async with session.post(
            "https://www.pge.com/eimpapi/auth/login",
            json={
                "username": username,
                "password": password,
                "appName": "CustomerSSO",
            },
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            result = await resp.json()
            if "errorMsg" in result:
                raise InvalidAuth(result["errorMsg"])

        url = (
            "https://itiamping.cloud.pge.com/idp/startSSO.ping?"
            "PartnerSpId=sso.opower.com&TargetResource="
        ) + urllib.parse.quote_plus(
            "https://pge.opower.com/ei/app/r/energy-usage-details"
        )

        await async_auth_saml(session, url)



"""Orange & Rockland Utilities (ORU)."""

from .coned import ConEd


class Oru(ConEd):
    """Orange & Rockland Utilities (ORU)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Orange & Rockland Utilities (ORU)"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "oru"

    @staticmethod
    def hostname() -> str:
        """Return the hostname for login."""
        return "oru.com"



"""Evergy."""

from html.parser import HTMLParser
import logging
from typing import Optional

import aiohttp

from ..const import USER_AGENT
from ..exceptions import InvalidAuth
from .base import UtilityBase

_LOGGER = logging.getLogger(__file__)


class EvergyLoginParser(HTMLParser):
    """HTML parser to extract login verification token from Evergy Login page."""

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self.verification_token: Optional[str] = None

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Try to extract the verification token from the login input."""
        if tag == "input" and ("name", "evrgaf") in attrs:
            _, token = next(filter(lambda attr: attr[0] == "value", attrs))
            self.verification_token = token


class Evergy(UtilityBase):
    """Evergy."""

    _subdomain: Optional[str] = None

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Evergy"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        assert Evergy._subdomain, "async_login not called"
        return Evergy._subdomain

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/Chicago"

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> str:
        """Login to the utility website."""
        login_parser = EvergyLoginParser()

        async with session.get(
            "https://www.evergy.com/log-in",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            login_parser.feed(await resp.text())

            assert (
                login_parser.verification_token
            ), "Failed to parse login verification token"

        login_payload = {
            "username": username,
            "password": password,
            "evrgaf": login_parser.verification_token,
        }

        async with session.post(
            "https://www.evergy.com/log-in",
            data=login_payload,
            headers={"User-Agent": USER_AGENT},
            allow_redirects=False,
            raise_for_status=False,
        ) as resp:
            # The response status will be 500 if verification token did not work
            if resp.status == 500:
                raise InvalidAuth("Login verification token failed")

            # Counterintuitively, if we get a 200 back, that means the login failed
            if resp.status == 200:
                raise InvalidAuth("Username and password failed")

        opower_access_token: Optional[str] = None

        async with session.get(
            "https://www.evergy.com/api/sso/jwt",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=False,
        ) as resp:
            opower_access_token = resp.headers["jwt"].removeprefix("Bearer ")

            assert opower_access_token, "Failed to parse OPower bearer token"

        async with session.get(
            "https://www.evergy.com/sc-api/account/getaccountpremiseselector",
            params={"isWidgetPage": "false", "hasNoSelector": "false"},
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            # returned mimetype is nonstandard, so this avoids a ContentTypeError
            data = await resp.json(content_type=None)
            # shape is: [{"accountNumber": 123456789, "oPowerDomain": "kcpl.opower.com", ...}]
            domain: str = data[0]["oPowerDomain"]
            Evergy._subdomain = domain.split(".", 1)[0]
            _LOGGER.debug("detected Evergy subdomain: %s", Evergy._subdomain)
            if Evergy._subdomain not in {"kcpk", "kcpl"}:
                _LOGGER.warn(
                    "unexpected Evergy subdomain %s, continuing",
                    Evergy._subdomain,
                )

        return opower_access_token



"""Public Service Company of Oklahoma (PSO)."""

from .aepbase import AEPBase
from .base import UtilityBase


class PSOklahoma(AEPBase, UtilityBase):
    """Public Service Company of Oklahoma (PSO)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Public Service Company of Oklahoma (PSO)"

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/Chicago"

    @staticmethod
    def hostname() -> str:
        """Return the hostname for login."""
        return "psoklahoma.com"



"""Indiana Michigan Power."""

from .aepbase import AEPBase
from .base import UtilityBase


class IndianaMichiganPower(AEPBase, UtilityBase):
    """Indiana Michigan Power."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Indiana Michigan Power"

    @staticmethod
    def hostname() -> str:
        """Return the hostname for login."""
        return "indianamichiganpower.com"



"""Base Abstract class for American Electric Power."""

from abc import ABC
from html.parser import HTMLParser
import re
from typing import Optional
import urllib.parse

import aiohttp

from ..const import USER_AGENT
from ..exceptions import InvalidAuth
from .helpers import async_auth_saml


class AEPLoginParser(HTMLParser):
    """HTML parser to extract login input fields."""

    def __init__(self, username: str, password: str) -> None:
        """Initialize."""
        super().__init__()
        self.inputs: dict[str, str] = {}
        self.username = username
        self.password = password
        self.password_field_found = False

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Try to extract the login input fields."""
        if tag == "input":
            name = ""
            value = ""
            for a in attrs:
                if a[0] == "name" and a[1] is not None:
                    name = a[1]
                if a[0] == "value" and a[1] is not None:
                    value = a[1]
            if "UserID" in name:
                value = self.username
            if "Password" in name:
                value = self.password
                self.password_field_found = True
            self.inputs[name] = value


class AEPTokenParser(HTMLParser):
    """HTML parser to extract token url and cookie."""

    _regexp = re.compile(r"var cookieKey = '(?P<cookieKey>\w+)'")

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self._in_inline_script = False
        self.cookieKey: Optional[str] = None
        self.token_url: Optional[str] = None

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Try to extract the token url."""
        if tag == "iframe":
            # Look for <iframe src=\"//www.aepohio.com/widgets/sso/opower?token=...&domain=www.aepohio.com\"
            for a in attrs:
                if a[0] == "src" and a[1] is not None and "opower" in a[1]:
                    self.token_url = a[1]

        # Recognizes inline scripts.
        if (
            tag == "script"
            and next(filter(lambda attr: attr[0] == "src", attrs), None) is None
        ):
            self._in_inline_script = True

    def handle_data(self, data: str) -> None:
        """Try to extract the cookie from the inline script."""
        if self._in_inline_script:
            result = self._regexp.search(data)
            if result and result.group("cookieKey"):
                self.cookieKey = result.group("cookieKey")

    def handle_endtag(self, tag: str) -> None:
        """Recognizes the end of inline scripts."""
        if tag == "script":
            self._in_inline_script = False


class AEPBase(ABC):
    """Base Abstract class for American Electric Power."""

    _subdomain: Optional[str] = None

    @classmethod
    def subdomain(cls) -> str:
        """Return the opower.com subdomain for this utility."""
        assert cls._subdomain, "async_login not called"
        return cls._subdomain

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/New_York"

    @staticmethod
    def hostname() -> str:
        """Return the hostname for login."""
        raise NotImplementedError

    @classmethod
    async def async_login(
        cls,
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> None:
        """Login in AEP using user/pass and then do the SAML call to opower."""
        # Clear cookies before logging in again, in case old ones are still around
        session.cookie_jar.clear(lambda c: c["domain"].endswith("opower.com"))

        login_parser = AEPLoginParser(username, password)
        token_parser = AEPTokenParser()

        # Get the login page and parse the ASP.Net Form Field that have generated names
        async with session.get(
            f"https://www.{cls.hostname()}/account/usage/",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            text = await resp.text()
            login_parser.feed(text)

        # Post the login page with the user credentials and get the cookieKey
        async with session.post(
            f"https://www.{cls.hostname()}/account/usage/",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
            data=login_parser.inputs,
        ) as resp:
            html = await resp.text()
            token_parser.feed(html)

        if token_parser.token_url is None or token_parser.cookieKey is None:
            raise InvalidAuth("Username/Password are invalid")

        match = re.search(r"https://([^.]*).opower.com", html)
        assert match
        cls._subdomain = match.group(1)

        url = (
            f"https://{cls.subdomain()}.opower.com/ei/x/embedded-api/authenticate?"
            + urllib.parse.urlencode(
                {
                    "client-url": f"https:{token_parser.token_url}&ou-session-initiated=true",
                    "error-param": "ou-auth-error",
                    "ou-entity-id": token_parser.cookieKey,
                }
            )
        )

        await async_auth_saml(session, url)



"""Puget Sound Energy (PSE)."""

from html.parser import HTMLParser
import re
from typing import Optional

import aiohttp

from ..const import USER_AGENT
from ..exceptions import InvalidAuth
from .base import UtilityBase


class PSELoginParser(HTMLParser):
    """HTML parser to extract login verification token from PSE Login page."""

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self.verification_token: Optional[str] = None

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Try to extract the verification token from the login input."""
        if tag == "input" and ("name", "__RequestVerificationToken") in attrs:
            _, token = next(filter(lambda attr: attr[0] == "value", attrs))
            self.verification_token = token


class PSEUsageParser(HTMLParser):
    """HTML parser to extract OPower bearer token from PSE Usage page."""

    _regexp = re.compile(r'var accessToken\s+=\s+["\'](?P<token>.+)["\']')

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self.opower_access_token: Optional[str] = None
        self._in_inline_script = False

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Recognizes inline scripts."""
        if (
            tag == "script"
            and next(filter(lambda attr: attr[0] == "src", attrs), None) is None
        ):
            self._in_inline_script = True

    def handle_data(self, data: str) -> None:
        """Try to extract the access token from the inline script."""
        if self._in_inline_script:
            result = self._regexp.search(data)
            if result and result.group("token"):
                self.opower_access_token = result.group("token")

    def handle_endtag(self, tag: str) -> None:
        """Recognizes the end of inline scripts."""
        if tag == "script":
            self._in_inline_script = False


class PSE(UtilityBase):
    """Puget Sound Energy (PSE)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Puget Sound Energy (PSE)"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "pse"

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/Los_Angeles"

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> str:
        """Login to the utility website."""
        # Double-logins are somewhat broken if cookies stay around.
        session.cookie_jar.clear(lambda cookie: cookie["domain"] == "www.pse.com")

        login_parser = PSELoginParser()

        async with session.get(
            "https://www.pse.com/en/login",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            login_parser.feed(await resp.text())

            assert (
                login_parser.verification_token
            ), "Failed to parse __RequestVerificationToken"

        await session.post(
            "https://www.pse.com/api/pseauthentication/AsyncSignIn",
            data={
                "__RequestVerificationToken": login_parser.verification_token,
                "GenericMessage": "Incorrect username or password",
                "LockedMessage": "Account is locked",
                "ReturnUrl": "",
                "OtherPageUrl": "False",
                "Username": username,
                "Password": password,
                "RememberMe": "true",
            },
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        )

        async with session.get(
            "https://www.pse.com/api/AccountSelector/GetContractAccountJson",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            if len(await resp.text()) == 0:
                raise InvalidAuth("Login failed")

        usage_parser = PSEUsageParser()

        async with session.get(
            "https://www.pse.com/en/account-and-billing/my-usage/view-my-usage",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            usage_parser.feed(await resp.text())

            assert (
                usage_parser.opower_access_token
            ), "Failed to parse OPower bearer token"

        return usage_parser.opower_access_token



"""Arizona Public Service (APS)."""

import logging
import re
from typing import Any, Optional

import aiohttp

from ..const import USER_AGENT
from ..exceptions import CannotConnect, InvalidAuth
from .base import UtilityBase
from .helpers import async_auth_saml, js_encrypt

_LOGGER = logging.getLogger(__name__)


class Aps(UtilityBase):
    """Arizona Public Service (APS)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Arizona Public Service (APS)"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "aps"

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/Phoenix"

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> None:
        """Login to the utility website."""
        _LOGGER.debug("Starting login process for Arizona Public Service (APS)")
        session.cookie_jar.clear(lambda cookie: cookie["domain"] == "www.aps.com")

        # Get public RSA key APS uses to encrypt the password
        async with session.get(
            "https://www.aps.com/Assets/Js/aps-apscom.js",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            html = await resp.text()
        rsa_key_text = extract_rsa_key(html)

        # Encrypt the password
        encrypted_password = js_encrypt(rsa_key_text, password)

        data = {
            "username": username,
            "password": encrypted_password,
        }

        # Send the POST request
        async with session.post(
            "https://www.aps.com/api/sitecore/SitecoreReactApi/UserAuthentication",
            json=data,
            raise_for_status=True,
        ) as resp:
            login_result = await resp.json(content_type="text/html")
            if login_result["isLoginSuccess"] is False:
                raise InvalidAuth("Username and password failed")

        # Get All User Details to get APS Account ID and Service Address ID
        async with session.get(
            "https://www.aps.com/api/sitecore/sitecorereactapi/GetAllUserDetails",
            raise_for_status=True,
        ) as resp:
            user_details = await resp.json(content_type="application /json")
            account_details = user_details["Details"]["AccountDetails"][
                "getAccountDetailsResponse"
            ]["getAccountDetailsRes"]
            account_id = account_details["getPersonDetails"]["accountID"]
            service_address_id = find_first_service_address_id(account_details)
            if service_address_id is None:
                raise CannotConnect("No APS Service Address ID found")

        account_service_id = f"{account_id}_{service_address_id}"

        # Start SAML authentication with APS and Opower
        url = f"https://www.aps.com/en/Residential/Save-Money-and-Energy/Opower?CA_SA={account_service_id}"
        try:
            await async_auth_saml(session, url)
        except aiohttp.ClientResponseError as e:
            if e.status == 403:
                raise CannotConnect(
                    "Forbidden error when authenticating with Opower. You might not have access to Opower."
                ) from e
            raise


def extract_rsa_key(js_content: str) -> str:
    """Extract the RSA public key from the APS JS file, using the identifier 'APSCOMWebPasswordpublicKey'."""
    pattern = r'APSCOMWebPasswordpublicKey:"(-----BEGIN PUBLIC KEY-----.*?-----END PUBLIC KEY-----)"'

    # Find the RSA key associated with the regex pattern above
    match = re.search(pattern, js_content, re.DOTALL)

    if match:
        # Get and format the RSA key
        rsa_key = match.group(1)
        formatted_key = re.sub(
            r"(-----BEGIN PUBLIC KEY-----)(.*)(-----END PUBLIC KEY-----)",
            r"\1\n\2\n\3",
            rsa_key,
            flags=re.DOTALL,
        )
        return formatted_key
    else:
        raise CannotConnect("The RSA public key was not found.")


def find_first_service_address_id(account_details: dict[str, Any]) -> Optional[str]:
    """Find the first service address ID from the account details."""
    try:
        premise_details_list = account_details["getSASPListByAccountID"][
            "premiseDetailsList"
        ]
        for premise in premise_details_list:
            sasp_details = premise.get("sASPDetails", [])
            for sasp in sasp_details:
                if "sAID" in sasp:
                    return str(sasp["sAID"])
    except (KeyError, TypeError):
        _LOGGER.warning(
            "Could not find APS Service Address ID in the expected structure"
        )
    return None



"""Base class for Exelon subsidiaries."""

import json
import logging
import re
from typing import Any, Optional

import aiohttp

from ..const import USER_AGENT
from ..exceptions import InvalidAuth

_LOGGER = logging.getLogger(__file__)


class Exelon:
    """Base class for Exelon subsidiaries."""

    _subdomain: Optional[str] = None

    # Can find the opower.com subdomain using the GetConfiguration endpoint
    # e.g. https://secure.bge.com/api/Services/MyAccountService.svc/GetConfiguration
    # returns bgec.opower.com

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/New_York"

    @staticmethod
    def login_domain() -> str:
        """Return the domain that hosts the login page."""
        raise NotImplementedError

    @classmethod
    def subdomain(cls) -> str:
        """Return the opower.com subdomain for this utility."""
        assert Exelon._subdomain, "async_login not called"
        return Exelon._subdomain

    @staticmethod
    def primary_subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        raise NotImplementedError

    @staticmethod
    def secondary_subdomain() -> str:
        """Return the opower.com secondary subdomain for this utility."""
        raise NotImplementedError

    @classmethod
    async def async_account(
        cls,
        session: aiohttp.ClientSession,
        bearer_token: str,
    ) -> Any:
        """Return the accounts for the current session."""
        # this path comes from GetConfiguration, unsure if its different
        # per utility, if so, would need to add it to the subtypes:
        # "euApiRoutePrefix": "/mobile/custom",
        eu_api_route_prefix = "/mobile/custom"
        # "euApiUrl": "/.euapi",
        eu_api_url = "/.euapi"
        async with session.get(
            "https://"
            + cls.login_domain()
            + eu_api_url
            + eu_api_route_prefix
            + "/auth/accounts",
            headers={
                "User-Agent": USER_AGENT,
                "Authorization": "Bearer " + bearer_token,
            },
            raise_for_status=True,
        ) as resp:
            # this has the wrong mime type for some reason
            result = await resp.json(content_type="text/html")

        if result["success"] is not True:
            raise InvalidAuth("Unable to list accounts")

        # Only include active accounts (after moving, old accounts have status: "Inactive")
        # NOTE: this logic currently assumes 1 active address per account, if multiple accounts found
        #      we default to taking the first in the list. Future enhancement is to support
        #      multiple accounts (which could result in different subdomain for each)
        active_accounts = [
            account for account in result["data"] if account["status"] == "Active"
        ]

        if len(active_accounts) == 0:
            raise InvalidAuth("No active accounts found")

        return active_accounts[0]

    @classmethod
    async def async_login(
        cls,
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> str:
        """Login to the utility website and authorize opower."""
        async with session.get(
            "https://" + cls.login_domain() + "/Pages/Login.aspx?/login",
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            result = await resp.text(encoding="utf-8")

        account = None
        # if we don't go to /accounts/dashboard, we need to perform some authorization steps
        if resp.request_info.url.path.endswith("/authorize"):
            # transId = "StateProperties=..."
            # policy = "B2C_1A_SignIn"
            # tenant = "/euazurebge.onmicrosoft.com/B2C_1A_SignIn"
            # api = "CombinedSigninAndSignup"
            settings_match = re.search(r"var SETTINGS = ({.*});", result)
            assert settings_match
            settings = json.loads(settings_match.group(1))
            login_post_domain = resp.real_url.host
            assert login_post_domain

            async with session.post(
                "https://"
                + login_post_domain
                + settings["hosts"]["tenant"]
                + "/SelfAsserted",
                params={
                    "tx": settings["transId"],
                    "p": settings["hosts"]["policy"],
                },
                data={
                    "request_type": "RESPONSE",
                    "signInName": username,
                    "password": password,
                },
                headers={
                    "X-CSRF-TOKEN": settings["csrf"],
                    "User-Agent": USER_AGENT,
                },
                raise_for_status=True,
            ) as resp:
                result_json = json.loads(await resp.text(encoding="utf-8"))

            if result_json["status"] != "200":
                raise InvalidAuth(result_json["message"])

            async with session.get(
                "https://"
                + login_post_domain
                + settings["hosts"]["tenant"]
                + "/api/"
                + settings["api"]
                + "/confirmed",
                params={
                    "rememberMe": settings["config"]["enableRememberMe"],
                    "csrf_token": settings["csrf"],
                    "tx": settings["transId"],
                    "p": settings["hosts"]["policy"],
                    "diags": json.dumps(
                        {
                            "pageViewId": settings["pageViewId"],
                            "pageId": settings["api"],
                            "trace": [],
                        }
                    ),
                },
                headers={"User-Agent": USER_AGENT},
                raise_for_status=True,
            ) as resp:
                result = await resp.text(encoding="utf-8")

            if resp.request_info.url.path.endswith(
                "/accounts/login/select-account"
            ) or resp.request_info.url.path.endswith("Pages/ChangeAccount.aspx"):
                # we probably need to select an account as we didn't automatically go to the dashboard
                async with session.get(
                    "https://"
                    + cls.login_domain()
                    + "/api/Services/MyAccountService.svc/GetSession",
                    headers={"User-Agent": USER_AGENT},
                    raise_for_status=True,
                ) as resp:
                    result_json = await resp.json()

                # confirm no account number is set
                if result_json["accountNumber"] is None:
                    bearer_token = result_json["token"]
                    # if we don't yet have an account, look one up and set it
                    if account is None:
                        account = await cls.async_account(session, bearer_token)

                    # set the first active one
                    account_number = account["accountNumber"]

                    async with session.post(
                        "https://"
                        + cls.login_domain()
                        + "/api/Services/AccountList.svc/ViewAccount",
                        json={
                            "accountNumber": account_number,
                        },
                        headers={"User-Agent": USER_AGENT},
                        raise_for_status=True,
                    ) as resp:
                        result = await resp.text(encoding="utf-8")

        async with session.post(
            "https://"
            + cls.login_domain()
            + "/api/Services/OpowerService.svc/GetOpowerToken",
            json={},
            headers={"User-Agent": USER_AGENT},
            raise_for_status=True,
        ) as resp:
            result_json = await resp.json()

        # If pepco or delmarva, determine if we should use secondary subdomain
        if cls.login_domain() in ["secure.pepco.com", "secure.delmarva.com"]:
            # Get the account type & state
            if account is None:
                account = await cls.async_account(session, result_json["access_token"])

            isResidential = account["isResidential"]
            state = account["PremiseInfo"][0]["mainAddress"]["townDetail"][
                "stateOrProvince"
            ]
            _LOGGER.debug("found exelon account isResidential: %s", isResidential)
            _LOGGER.debug("found exelon account state: %s", state)

            # Determine subdomain to use by matching logic found in https://cls.login_domain()/dist/app.js
            Exelon._subdomain = cls.primary_subdomain()
            if not isResidential or state != "MD":
                Exelon._subdomain = cls.secondary_subdomain()

            _LOGGER.debug("detected exelon subdomain to be: %s", Exelon._subdomain)

        return str(result_json["access_token"])



"""OIDC Login Helper and its constituent functions."""

import asyncio
import base64
import hashlib
import json
import logging
import secrets
import ssl
from typing import Any, Optional, TypedDict
from urllib.parse import parse_qs, urlparse

import aiohttp

from ..exceptions import CannotConnect, InvalidAuth

_LOGGER = logging.getLogger(__name__)


async def async_auth_oidc(
    username: str,
    password: str,
    base_url: str,
    tenant_id: str,
    policy: str,
    client_id: str,
    redirect_uri: str,
    scope_auth: str,
    scope_access: str,
    self_asserted_endpoint: str,
    policy_confirm_endpoint: str,
) -> Optional[str]:
    """Perform the login process and return an access token."""
    ssl_context = await asyncio.get_running_loop().run_in_executor(
        None, ssl.create_default_context
    )
    connector = aiohttp.TCPConnector(ssl=ssl_context)
    secure_session = aiohttp.ClientSession(connector=connector)
    try:
        code_verifier = _generate_code_verifier()
        code_challenge = _generate_code_challenge(code_verifier)
        _LOGGER.debug("Generated PKCE code verifier and challenge")
        config = await _get_config(secure_session, base_url, tenant_id, policy)
        _LOGGER.debug("Retrieved OAuth configuration")
        auth_code = await _get_auth(
            secure_session,
            config,
            code_challenge,
            username,
            password,
            client_id,
            redirect_uri,
            scope_auth,
            policy,
            self_asserted_endpoint,
            policy_confirm_endpoint,
        )
        if auth_code is None:
            _LOGGER.error("Failed to obtain authorization code")
            raise CannotConnect("Failed to obtain authorization code")
        _LOGGER.debug("Obtained authorization code")

        tokens = await _get_access(
            secure_session,
            config,
            auth_code,
            code_verifier,
            client_id,
            redirect_uri,
            scope_access,
        )

        if tokens and "access_token" in tokens:
            _LOGGER.debug("Successfully obtained access token")
            return tokens["access_token"]
        else:
            _LOGGER.error("Failed to obtain access token")
            raise CannotConnect("Failed to obtain access token")

    except aiohttp.ClientError as err:
        _LOGGER.error("Connection error during login: %s", str(err))
        raise CannotConnect(f"Connection error: {err}")
    finally:
        await secure_session.close()


class ConfigDict(TypedDict):
    """Dictionary to store configuration details for OAuth."""

    authorization_endpoint: str
    issuer: str
    token_endpoint: str


class TokenDict(TypedDict):
    """Dictionary to store OAuth tokens."""

    access_token: str


def _generate_code_verifier() -> str:
    """Generate a code verifier for PKCE."""
    return secrets.token_urlsafe(32)


def _generate_code_challenge(code_verifier: str) -> str:
    """Generate a code challenge for PKCE."""
    code_challenge_digest = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    return base64.urlsafe_b64encode(code_challenge_digest).decode("utf-8").rstrip("=")


async def _get_config(
    session: aiohttp.ClientSession, base_url: str, tenant_id: str, policy: str
) -> ConfigDict:
    """Get the configuration from the server."""
    config_url = (
        f"{base_url}/{tenant_id}/{policy}/v2.0/.well-known/openid-configuration"
    )
    _LOGGER.debug("Fetching OAuth configuration from: %s", config_url)
    config_text, _, status = await _fetch(session, config_url)
    if status != 200 or not config_text:
        _LOGGER.error("Failed to get configuration. Status: %s", status)
        raise CannotConnect("Failed to get configuration")
    config: ConfigDict = json.loads(config_text)
    return config


async def _get_auth(
    session: aiohttp.ClientSession,
    config: ConfigDict,
    code_challenge: str,
    username: str,
    password: str,
    client_id: str,
    redirect_uri: str,
    scope_auth: str,
    policy: str,
    self_asserted_endpoint: str,
    policy_confirm_endpoint: str,
) -> Optional[str]:
    """Get the authorization code."""
    auth_params = {
        "client_id": client_id,
        "response_type": "code",
        "redirect_uri": redirect_uri,
        "scope": scope_auth,
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
    }
    _LOGGER.debug("Requesting authorization code")
    auth_content, final_url, status = await _fetch(
        session, config["authorization_endpoint"], params=auth_params
    )
    if status != 200 or not auth_content:
        _LOGGER.error("Failed to get authorization. Status: %s", status)
        raise CannotConnect("Failed to get authorization")

    settings = _extract_settings(auth_content)
    if not settings:
        _LOGGER.debug("No settings extracted, checking for direct authorization code")
        if final_url and final_url.startswith(redirect_uri):
            query = urlparse(final_url).query
            parsed_query = parse_qs(query)
            return parsed_query.get("code", [None])[0]
        return None

    _LOGGER.debug("Posting credentials")
    await _post_credentials(
        session,
        config["issuer"],
        settings,
        username,
        password,
        policy,
        self_asserted_endpoint,
    )
    _LOGGER.debug("Confirming sign-in")
    return await _confirm_signin(
        session, config["issuer"], settings, policy, policy_confirm_endpoint
    )


async def _get_access(
    session: aiohttp.ClientSession,
    config: ConfigDict,
    auth_code: str,
    code_verifier: str,
    client_id: str,
    redirect_uri: str,
    scope_access: str,
) -> Optional[TokenDict]:
    """Get the access token."""
    token_data = {
        "client_id": client_id,
        "grant_type": "authorization_code",
        "code": auth_code,
        "redirect_uri": redirect_uri,
        "code_verifier": code_verifier,
        "scope": scope_access,
    }
    _LOGGER.debug("Requesting access token")
    token_content, _, status = await _fetch(
        session, config["token_endpoint"], method="POST", data=token_data
    )
    if status != 200 or not token_content:
        _LOGGER.error("Failed to get access token. Status: %s", status)
        raise CannotConnect("Failed to get access token")
    tokens: TokenDict = json.loads(token_content)
    return tokens


async def _fetch(
    session: aiohttp.ClientSession, url: str, **kwargs: Any
) -> tuple[Optional[str], Optional[str], int]:
    """Fetch data from a URL."""
    method = kwargs.pop("method", "GET")
    timeout = aiohttp.ClientTimeout(total=30)
    try:
        _LOGGER.debug("Fetching URL: %s, Method: %s", url, method)
        async with session.request(method, url, timeout=timeout, **kwargs) as response:
            content = await response.text()
            _LOGGER.debug("Fetch completed. Status: %s", response.status)
            return content, str(response.url), response.status
    except aiohttp.ClientError as e:
        _LOGGER.error("Network error occurred: %s", str(e))
        return None, None, 0


def _extract_settings(auth_content: str) -> Optional[dict[str, Any]]:
    """Extract settings from the authorization content."""
    _LOGGER.debug("Extracting settings from authorization content")
    settings_start = auth_content.find("var SETTINGS = ")
    if settings_start == -1:
        _LOGGER.debug("Settings not found in authorization content")
        return None
    settings_end = auth_content.find(";", settings_start)
    if settings_end == -1:
        _LOGGER.debug("End of settings not found in authorization content")
        return None
    settings_json = auth_content[settings_start + 15 : settings_end].strip()
    try:
        settings: dict[str, Any] = json.loads(settings_json)
        _LOGGER.debug("Settings successfully extracted")
        return settings
    except json.JSONDecodeError:
        _LOGGER.error("Failed to parse settings JSON")
        return None


async def _post_credentials(
    session: aiohttp.ClientSession,
    issuer: str,
    settings: dict[str, Any],
    username: str,
    password: str,
    policy: str,
    self_asserted_endpoint: str,
) -> None:
    """Post credentials to the server."""
    base_url = issuer.rsplit("/", 2)[0]
    _LOGGER.debug("Posting credentials to %s", base_url)
    _, _, status = await _fetch(
        session,
        f"{base_url}/{policy}/{self_asserted_endpoint}",
        method="POST",
        data={
            "tx": settings["transId"],
            "p": policy,
            "request_type": "RESPONSE",
            "signInName": username,
            "password": password,
        },
        headers={"X-CSRF-TOKEN": settings["csrf"]},
    )
    if status != 200:
        _LOGGER.error("Failed to post credentials. Status: %s", status)
        raise InvalidAuth("Invalid username or password")
    _LOGGER.debug("Credentials posted successfully")


async def _confirm_signin(
    session: aiohttp.ClientSession,
    issuer: str,
    settings: dict[str, Any],
    policy: str,
    policy_confirm_endpoint: str,
) -> Optional[str]:
    """Confirm the sign-in process."""
    base_url = issuer.rsplit("/", 2)[0]
    _LOGGER.debug("Confirming sign-in at %s", base_url)
    _, final_url, status = await _fetch(
        session,
        f"{base_url}/{policy}/{policy_confirm_endpoint}",
        params={
            "rememberMe": "false",
            "csrf_token": settings["csrf"],
            "tx": settings["transId"],
            "p": policy,
        },
        allow_redirects=True,
    )
    if status != 200:
        _LOGGER.error("Failed to confirm signin. Status: %s", status)
        raise CannotConnect("Failed to confirm signin")
    if final_url:
        query = urlparse(final_url).query
        parsed_query = parse_qs(query)
        auth_code = parsed_query.get("code", [None])[0]
        if auth_code:
            _LOGGER.debug("Sign-in confirmed, authorization code obtained")
        else:
            _LOGGER.warning("Sign-in confirmed, but no authorization code found")
        return auth_code
    _LOGGER.warning("Sign-in confirmation did not result in a final URL")
    return None



"""Demo usage of Opower library."""

import argparse
import asyncio
import csv
from datetime import datetime, timedelta
from getpass import getpass
import logging
from typing import Optional

import aiohttp

from opower import (
    AggregateType,
    Opower,
    ReadResolution,
    get_supported_utilities,
    select_utility,
)


async def _main() -> None:
    supported_utilities = [
        utility.__name__.lower() for utility in get_supported_utilities()
    ]
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--utility",
        help="Utility. If not provided, you will be asked for it",
        choices=supported_utilities,
        type=str.lower,
    )
    parser.add_argument(
        "--username",
        help="Username for logging into the utility's website. "
        "If not provided, you will be asked for it",
    )
    parser.add_argument(
        "--password",
        help="Password for logging into the utility's website. "
        "If not provided, you will be asked for it",
    )
    parser.add_argument(
        "--mfa_secret",
        help="MFA secret for logging into the utility's website.",
    )
    parser.add_argument(
        "--aggregate_type",
        help="How to aggregate historical data. Defaults to day",
        choices=list(AggregateType),
        type=AggregateType,
        default=AggregateType.DAY,
    )
    parser.add_argument(
        "--start_date",
        help="Start datetime for historical data. Defaults to 7 days ago",
        type=lambda s: datetime.fromisoformat(s),
        default=datetime.now() - timedelta(days=7),
    )
    parser.add_argument(
        "--end_date",
        help="end datetime for historical data. Defaults to now",
        type=lambda s: datetime.fromisoformat(s),
        default=datetime.now(),
    )
    parser.add_argument(
        "--usage_only",
        help="If true will output usage only, not cost",
        action="store_true",
    )
    parser.add_argument(
        "--csv",
        help="csv file to store data",
    )
    parser.add_argument(
        "-v", "--verbose", help="enable verbose logging", action="count", default=0
    )
    parser.add_argument(
        "--realtime",
        help="If true, fetches usage-only data from the realtime API. "
        "Not all utilities support the realtime API.",
        action="store_true",
    )
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG - args.verbose + 1 if args.verbose > 0 else logging.INFO
    )

    utility = args.utility or input(f"Utility, one of {supported_utilities}: ")
    username = args.username or input("Username: ")
    password = args.password or getpass("Password: ")
    mfa_secret = args.mfa_secret or (
        input("2FA secret: ") if select_utility(utility).accepts_mfa() else None
    )

    async with aiohttp.ClientSession() as session:
        opower = Opower(session, utility, username, password, mfa_secret)
        await opower.async_login()
        if not args.csv:
            # Re-login to make sure code handles already logged in sessions.
            await opower.async_login()
            for forecast in await opower.async_get_forecast():
                print("\nCurrent bill forecast:", forecast)
        for account in await opower.async_get_accounts():
            aggregate_type = args.aggregate_type
            if (
                aggregate_type == AggregateType.HOUR
                and account.read_resolution == ReadResolution.DAY
            ):
                aggregate_type = AggregateType.DAY
            elif (
                aggregate_type != AggregateType.BILL
                and account.read_resolution == ReadResolution.BILLING
            ):
                aggregate_type = AggregateType.BILL
            if not args.csv:
                print(
                    "\nGetting historical data: account=",
                    account,
                    "aggregate_type=",
                    aggregate_type,
                    "start_date=",
                    args.start_date,
                    "end_date=",
                    args.end_date,
                )
            prev_end: Optional[datetime] = None
            # Realtime data does not include cost data, so effectively --realtime implies --usage_only.
            if args.usage_only or args.realtime:
                if args.realtime:
                    usage_data = await opower.async_get_realtime_usage_reads(account)
                else:
                    usage_data = await opower.async_get_usage_reads(
                        account,
                        aggregate_type,
                        args.start_date,
                        args.end_date,
                    )
                if args.csv:
                    with open(args.csv, "w", newline="") as csv_file:
                        writer = csv.writer(csv_file)
                        writer.writerow(["start_time", "end_time", "consumption"])
                        for usage_read in usage_data:
                            writer.writerow(
                                [
                                    usage_read.start_time,
                                    usage_read.end_time,
                                    usage_read.consumption,
                                ]
                            )
                else:
                    print(
                        "start_time\tend_time\tconsumption"
                        "\tstart_minus_prev_end\tend_minus_prev_end"
                    )
                    for usage_read in usage_data:
                        start_minus_prev_end = (
                            None
                            if prev_end is None
                            else usage_read.start_time - prev_end
                        )
                        end_minus_prev_end = (
                            None if prev_end is None else usage_read.end_time - prev_end
                        )
                        prev_end = usage_read.end_time
                        print(
                            f"{usage_read.start_time}"
                            f"\t{usage_read.end_time}"
                            f"\t{usage_read.consumption}"
                            f"\t{start_minus_prev_end}"
                            f"\t{end_minus_prev_end}"
                        )
                    print()
            else:
                cost_data = await opower.async_get_cost_reads(
                    account,
                    aggregate_type,
                    args.start_date,
                    args.end_date,
                )
                if args.csv:
                    with open(args.csv, "w", newline="") as csv_file:
                        writer = csv.writer(csv_file)
                        writer.writerow(
                            ["start_time", "end_time", "consumption", "provided_cost"]
                        )
                        for cost_read in cost_data:
                            writer.writerow(
                                [
                                    cost_read.start_time,
                                    cost_read.end_time,
                                    cost_read.consumption,
                                    cost_read.provided_cost,
                                ]
                            )
                else:
                    print(
                        "start_time\tend_time\tconsumption\tprovided_cost"
                        "\tstart_minus_prev_end\tend_minus_prev_end"
                    )
                    for cost_read in cost_data:
                        start_minus_prev_end = (
                            None
                            if prev_end is None
                            else cost_read.start_time - prev_end
                        )
                        end_minus_prev_end = (
                            None if prev_end is None else cost_read.end_time - prev_end
                        )
                        prev_end = cost_read.end_time
                        print(
                            f"{cost_read.start_time}"
                            f"\t{cost_read.end_time}"
                            f"\t{cost_read.consumption}"
                            f"\t{cost_read.provided_cost}"
                            f"\t{start_minus_prev_end}"
                            f"\t{end_minus_prev_end}"
                        )
                    print()


if __name__ == "__main__":
    asyncio.run(_main())



"""PECO Energy Company (PECO)."""

from .base import UtilityBase
from .exelon import Exelon


class PECO(Exelon, UtilityBase):
    """PECO Energy Company (PECO)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "PECO Energy Company (PECO)"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "peco"

    @staticmethod
    def login_domain() -> str:
        """Return the domain that hosts the login page."""
        return "secure.peco.com"



"""Mercury NZ Limited utility implementation.

This module handles the authentication and API interactions for Mercury NZ.
It uses OAuth 2.0 with PKCE for secure authentication.
"""

import logging
from typing import Optional

import aiohttp

from .base import UtilityBase
from .oidchelper import async_auth_oidc

_LOGGER = logging.getLogger(__name__)


class Mercury(UtilityBase):
    """Mercury NZ Limited utility implementation.

    This class handles the authentication and API interactions for Mercury NZ.
    It uses OAuth 2.0 with PKCE for secure authentication.
    """

    BASE_URL = "https://login.mercury.co.nz"
    TENANT_ID = "fc07dca7-cd6a-4578-952b-de7a7afaebdc"
    POLICY = "b2c_1a_signup_signin"
    CLIENT_ID = "4c8c2c47-24cd-485d-aad9-12f3d95b3ceb"
    REDIRECT_URI = "https://myaccount.mercury.co.nz"
    APPLICATION_URI = f"{BASE_URL}/aded9884-533e-4081-a4ce-87b0d4e80a45/"
    SCOPE_AUTH = "openid profile offline_access"
    SCOPE_ACCESS = f"{APPLICATION_URI}customer:write {APPLICATION_URI}customer:read openid profile offline_access"
    SELF_ASSERTED_ENDPOINT = "SelfAsserted"
    POLICY_CONFIRM_ENDPOINT = "api/CombinedSigninAndSignup/confirmed"

    @staticmethod
    def name() -> str:
        """Return the name of the utility."""
        return "Mercury NZ Limited"

    @staticmethod
    def subdomain() -> str:
        """Return the subdomain for the utility."""
        return "mercury"

    @staticmethod
    def utilitycode() -> str:
        """Return the utilitycode identifier for the utility."""
        return "meen"

    @staticmethod
    def timezone() -> str:
        """Return the timezone for the utility."""
        return "Pacific/Auckland"

    @staticmethod
    def is_dss() -> bool:
        """Check if the utility uses DSS version of the portal."""
        return False

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> Optional[str]:
        """Perform the login process and return an access token."""
        _LOGGER.debug("Starting login process for Mercury NZ Limited")
        return await async_auth_oidc(
            username,
            password,
            Mercury.BASE_URL,
            Mercury.TENANT_ID,
            Mercury.POLICY,
            Mercury.CLIENT_ID,
            Mercury.REDIRECT_URI,
            Mercury.SCOPE_AUTH,
            Mercury.SCOPE_ACCESS,
            Mercury.SELF_ASSERTED_ENDPOINT,
            Mercury.POLICY_CONFIRM_ENDPOINT,
        )



"""Duquesne Light Company (DQE)."""

from html.parser import HTMLParser
import re
from typing import Optional

import aiohttp

from ..const import USER_AGENT
from ..exceptions import InvalidAuth
from .base import UtilityBase


class DQEUsageParser(HTMLParser):
    """HTML parser to extract OPower bearer token from DQE Usage page."""

    _regexp = re.compile(r'"OPowerToken": "(?P<token>.+)"')

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self.opower_access_token: Optional[str] = None
        self._in_inline_script = False

    def handle_starttag(self, tag: str, attrs: list[tuple[str, Optional[str]]]) -> None:
        """Recognizes inline scripts."""
        if (
            tag == "script"
            and next(filter(lambda attr: attr[0] == "src", attrs), None) is None
        ):
            self._in_inline_script = True

    def handle_data(self, data: str) -> None:
        """Try to extract the access token from the inline script."""
        if self._in_inline_script:
            result = self._regexp.search(data)
            if result and result.group("token"):
                self.opower_access_token = result.group("token")

    def handle_endtag(self, tag: str) -> None:
        """Recognizes the end of inline scripts."""
        if tag == "script":
            self._in_inline_script = False


class DuquesneLight(UtilityBase):
    """Duquesne Light Company (DQE)."""

    @staticmethod
    def name() -> str:
        """Distinct recognizable name of the utility."""
        return "Duquesne Light Company (DQE)"

    @staticmethod
    def subdomain() -> str:
        """Return the opower.com subdomain for this utility."""
        return "duq"

    @staticmethod
    def timezone() -> str:
        """Return the timezone."""
        return "America/New_York"

    @staticmethod
    async def async_login(
        session: aiohttp.ClientSession,
        username: str,
        password: str,
        optional_mfa_secret: Optional[str],
    ) -> str:
        """Login to the utility website."""
        # Double-logins are somewhat broken if cookies stay around.
        session.cookie_jar.clear(
            lambda cookie: cookie["domain"] == "www.duquesnelight.com"
        )
        # DQE uses Incapsula and merely passing the User-Agent is not enough.
        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "Sec-Fetch-Dest": "document",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-Site": "none",
            "Sec-Fetch-User": "?1",
            "Cache-Control": "max-age=0",
        }

        async with session.post(
            "https://www.duquesnelight.com/login/login",
            data={
                "Phone": "",
                "Email": "",
                "IsLoginOff": "false",
                "RedFlagPassword": "",
                "RememberUsername": "false",
                "Username": username,
                "Password": password,
                "RedirectUrl": "/my-account/account-summary",
                "SuppressPleaseLoginMessage": "true",
                "LoginTurnedOffMessage": "",
                "RedirectPath": "",
                "PersonId": "",
            },
            headers=headers,
            raise_for_status=True,
        ) as resp:
            # Check for failed login - DQE returns status 200 with a json body that can be parsed.
            if "invalid" in await resp.text():
                raise InvalidAuth("Login failed")

        usage_parser = DQEUsageParser()

        async with session.get(
            "https://www.duquesnelight.com/energy-money-savings/my-electric-use",
            headers=headers,
            raise_for_status=True,
        ) as resp:
            usage_parser.feed(await resp.text())

            assert (
                usage_parser.opower_access_token
            ), "Failed to parse OPower bearer token"

        return usage_parser.opower_access_token
